#include "Acquisition.hpp"
#include "FreeRTOS.h"
#include "task.h"
#include "PPI.h"
#include "ports.h"
#include "HSIS.h"
#include "NVIC.h"
#include "easyDAC.hpp"
#include <cstring> // memcpy
#include "GPIO.h"
#include "cough.hpp"
#include "TRH.hpp"
#include "DPT.hpp"
#include "neb.hpp"
#include "IRQ_Priority.h"
#include "Task_Priority.h"
#include "Implant.hpp"

// Preprocessor stuff / magic numbers
#define STACK_SIZE 100
#define Acq_Count_2kHz 1024
#define Acq_Count_1kHz 2048
#define Acq_Count_500Hz 4096
#define Acq_Count_250Hz 8192
#define Acq_Count_125Hz 16384
#define QUERY_UNLOCK_MS 100
#define Signals_Count 12U // 4 vivoFlow + 8 easyTEL+

// Read only const
enum{
	DPT,
	Temp,
	Humidity,
	Cough,
	Implant_iFast1,
	Implant_iFast2,
	Implant_iFast3,
	Implant_iFast4,
	Implant_iT,
	Implant_iAcc,
	Implant_iVbat,
	Implant_QoS
};

// not static because used by easyDAC class
bool IsAcqEnabled; 
vF_Acq_Frame Acquisition_Buffer;

// Static local vars
static TaskHandle_t AcqTaskHandle;
static StaticTask_t xTaskBuffer;
static StackType_t xStack[STACK_SIZE];
static TaskHandle_t QueryTaskHandle;
static StaticTask_t xTaskBuffer2;
static StackType_t xStack2[STACK_SIZE];
static uint8_t AcqTickWrapValue = 7;
static unsigned char SignalsFreq[Signals_Count]; // 1 = 125Hz, 2 = 250Hz, etc
static unsigned char SignalsPeriod[Signals_Count][16u]; // 8 value is related to freq format (n*125Hz) sent by master
static unsigned char Acq_Tick = 0;

// Methods

void Acquisition::Setup()
{
	IsAcqEnabled = false;

	GPIOTE_Config gpiote_cfg_no_irq = {
		.Mode = GPIOTE_Event,
		.Polarity = GPIOTE_LowToHigh,
		.Channel = Acquisition_GPIOTE_Channel,
		.Pin = EZDAC_CLK_Pin,
		.Port = EZDAC_CLK_Port,
		.Event_Callback = NULL // unused with no irq setup
	};
	GPIOTE_Setup_NO_IRQ(gpiote_cfg_no_irq);

	PPI_Config ppi = {
		.event = GPIOTE_EVENTS_IN_0,
		.task = TIMER0_TASKS_COUNT,
		.channel_index = Acquisition_PPI_Channel
	};
	PPI_Configure_Channel(ppi);
	PPI_Enable_Channel(Acquisition_PPI_Channel);

	IRQ_SET(IRQ_Timer0, ezDAC_CLK_PRIO);
	IRQ_SET(IRQ_Serial0, DPT_TRH_SER_PRIO);
	IRQ_SET(IRQ_GPIOTE, GPIOTE_PRIO);
	IRQ_SET(IRQ_ADC, ADC_PRIO);

	Cough::Setup();
	DPT::Setup(&QueryTaskHandle);
	Neb::Setup();

	AcqTaskHandle = xTaskCreateStatic(Acq_Task, "ACQ", STACK_SIZE, NULL, ACQ_PRIORITY, xStack, &xTaskBuffer);
	QueryTaskHandle = xTaskCreateStatic(Query_Task, "DPT", STACK_SIZE, NULL, QUERY_PRIORITY, xStack2, &xTaskBuffer2);
}

void Acquisition::Acq_Task(void* context)
{
	
	unsigned char bulk_index = 2;
	(void)context;
	Acquisition_Buffer.vF_status = 0x11;
	while(1)
	{
		(void)ulTaskNotifyTake(pdTRUE, portMAX_DELAY); // wakes at Fmax created from master supplied 2.048MHz clock
		memset(Acquisition_Buffer.bulk + 2, 0, sizeof(Acquisition_Buffer.bulk) - 2);
		bulk_index = 2; // why 2 and not 0 ? Because frame starts with 2 status bytes
		for(unsigned int i = 0; i < Signals_Count; i++)
			if(SignalsPeriod[i][Acq_Tick]) // Is it time to replace the old sample with a new sample
				switch(i)
				{
					case DPT:
					DPT::GetSample(&Acquisition_Buffer.bulk[bulk_index]);
					bulk_index+=3;
					break;
					case Temp:
					TRH::GetTemperatureSample(&Acquisition_Buffer.bulk[bulk_index]);
					bulk_index+=2;
					break;
					case Humidity:
					TRH::GetHumiditySample(&Acquisition_Buffer.bulk[bulk_index]);
					bulk_index+=2;
					break;
					case Cough:
					Cough::GetSample(&Acquisition_Buffer.bulk[bulk_index]);
					bulk_index+=2;
					break;
					case Implant_iFast1:
					TImplant::GetSample(&Acquisition_Buffer.bulk[bulk_index], iFast1);
					bulk_index+=2;
					break;
					case Implant_iFast2:
					TImplant::GetSample(&Acquisition_Buffer.bulk[bulk_index], iFast2);
					bulk_index+=3;
					break;
					case Implant_iFast3:
					TImplant::GetSample(&Acquisition_Buffer.bulk[bulk_index], iFast3);
					bulk_index+=2;
					break;
					case Implant_iFast4:
					TImplant::GetSample(&Acquisition_Buffer.bulk[bulk_index], iFast4);
					bulk_index+=2;
					break;
					case Implant_iT:
					TImplant::GetSample(&Acquisition_Buffer.bulk[bulk_index], iT);
					bulk_index+=2;
					break;
					case Implant_iAcc:
					TImplant::GetSample(&Acquisition_Buffer.bulk[bulk_index], iAcc);
					bulk_index+=2;
					break;
					case Implant_iVbat:
					TImplant::GetSample(&Acquisition_Buffer.bulk[bulk_index], iVbat);
					bulk_index+=2;
					break;
					case Implant_QoS:
					TImplant::GetSample(&Acquisition_Buffer.bulk[bulk_index], QoS);
					bulk_index+=2;
					break;
				}
		easyDAC::Prepare_Acq_Frame(bulk_index); // Make sure slave is aiming at the right buffers
		io_clr(TRIG_Port, TRIG_Pin); // Issue a falling edge on TRIG to signal master samples have been gathered
		Acq_Tick = Acq_Tick == AcqTickWrapValue ? 0 : Acq_Tick + 1; // Update tick
	}
}

// This tasks is dedicated to constant readings of the DPT and TRH sensors (using a shared port)
void Acquisition::Query_Task(void* context) 
{
	static unsigned int loop_count = 0;
	(void)context;
	(void)DPT::StartBlocking();
	while(1)
	{
		switch(loop_count++)
		{
			case 0UL:
			if(IsAcqEnabled && TRH::IsDetected()) TRH::Start();
			break;
			case 124UL: // Multiply by 4 ms to get approx period
			if(IsAcqEnabled && TRH::IsDetected()) TRH::ReadData();
			break;
			case 249UL:
			loop_count = 0;
			break;
			default:
			break;
		}
		(void)ulTaskNotifyTake(pdTRUE, QUERY_UNLOCK_MS); // Wait to be awaken
		if(IsAcqEnabled)
		{
			(void)DPT::ReadDataBlocking();
			(void)DPT::StartBlocking();
		}
	}
}

void Acquisition::Detect(unsigned char* state)
{
	unsigned char new_state = 0;
	// Temp and RH detected
	if(TRH::Detect()) new_state += 1;
	// Neb detected
	if(Neb::Detect()) new_state += 2;
	state[0] = new_state;
}

void Acquisition::Disable()
{
	Cough::Stop();
	IsAcqEnabled = false;
}

// Set period for config highest freq (Max freq = 2kHz)
static void Set_Sig_Period(unsigned char* signal_period, unsigned char signal_freq, unsigned char HighestFreq)
{
	const unsigned char ACQ_ARRAY_0[]= 		{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	const unsigned char ACQ_ARRAY_1[]= 		{1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1};
	const unsigned char ACQ_ARRAY_1_2[]= 	{1,0,1,0,1,0,1,0,1,0,1,0,1,0,1,0};
	const unsigned char ACQ_ARRAY_1_4[]= 	{1,0,0,0,1,0,0,0,1,0,0,0,1,0,0,0};
	const unsigned char ACQ_ARRAY_1_8[]= 	{1,0,0,0,0,0,0,0,1,0,0,0,0,0,0,0};
	const unsigned char ACQ_ARRAY_1_16[]= 	{1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
	if(signal_freq == 0) 
	{
		(void)memcpy(signal_period, ACQ_ARRAY_0, sizeof(ACQ_ARRAY_0));
		return;
	}
	unsigned char Freq_Ratio = HighestFreq / signal_freq;
	const unsigned char* acq_array = ACQ_ARRAY_0;
	if(Freq_Ratio == 1) acq_array = ACQ_ARRAY_1;
	else if (Freq_Ratio == 2) acq_array = ACQ_ARRAY_1_2;
	else if (Freq_Ratio == 4) acq_array = ACQ_ARRAY_1_4;
	else if (Freq_Ratio == 8) acq_array = ACQ_ARRAY_1_8;
	else if (Freq_Ratio == 16) acq_array = ACQ_ARRAY_1_16;
	(void)memcpy(signal_period, acq_array, sizeof(ACQ_ARRAY_0));
}

enum{
	freq_125Hz = 1,
	freq_250Hz = 2,
	freq_500Hz = 4,
	freq_1kHz = 8,
	freq_2kHz = 16
};

// Set counter value for new acq period
static void Set_Acq_Period(unsigned char HighestFreq)
{
	uint32_t Regval=0;
	switch(HighestFreq)
	{
		case freq_2kHz:
		Regval = Acq_Count_2kHz;
		break;
		case freq_1kHz:
		Regval = Acq_Count_1kHz;
		break;
		case freq_500Hz:
		Regval = Acq_Count_500Hz;
		break;
		case freq_250Hz:
		Regval = Acq_Count_250Hz;
		break;
		case freq_125Hz:
		Regval = Acq_Count_125Hz;
		break;
	}
	Counter_Config count_cfg = {
			.Channel = Acquisition_Timer_Channel,
			.Bitmode = Bits16,
			.Compare_Channel = Acquisition_Timer_Compare_Channel,
			.CompareValue = Regval,
			.Callback = Acquisition::OnTick
		};
	Counter_Setup(count_cfg);
}

void Acquisition::Enable()
{
	bool ImplantSignalInAcq = false;
	unsigned char HighestFreq = 0;
	if(!IsAcqEnabled)
	{
		unsigned char* ezDAC_rxed_freqs = easyDAC::GetSignalsConfig();
		// Populate SignalsFreq with ezDAC_rxed_freqs
		for(unsigned int i = 0; i < Signals_Count; i++) SignalsFreq[i] = ezDAC_rxed_freqs[i];
		// Find HighestFreq with SignalsFreq
		for(unsigned int i = 0; i < Signals_Count; i++) HighestFreq = SignalsFreq[i] > HighestFreq ? SignalsFreq[i] : HighestFreq;
		// Populate SignalsPeriod with SignalsFreq and HighestFreq
		for(unsigned int i = 0; i < Signals_Count; i++) Set_Sig_Period(SignalsPeriod[i], SignalsFreq[i], HighestFreq);
		// Update acquisition wraping value for new acq freq
		AcqTickWrapValue = HighestFreq - 1;
		Acq_Tick = 0;
		Set_Acq_Period(HighestFreq);
		Cough::Start();

		// Check if implant signal are in use
		for(unsigned int i = Implant_iFast1; i < Signals_Count; i++)
			if(SignalsFreq[i] > 0) ImplantSignalInAcq = true;
		
		if(ImplantSignalInAcq && easyDAC::ImplantConfigReceived)
			TImplant::StartConfig(easyDAC::GetImplantConfig());
		//else TImplant::StartConfig(easyDAC::GetImplantConfig());
		IsAcqEnabled = true;
	}
}

void Acquisition::OnTick()
{
	BaseType_t xHigherPriorityTaskWoken = pdFALSE;
	if(IsAcqEnabled)
	{
		vTaskNotifyGiveFromISR(AcqTaskHandle, &xHigherPriorityTaskWoken);
		portYIELD_FROM_ISR(xHigherPriorityTaskWoken);
	}
}
