// Radio driver for nRF52840 | gat 11/04/23
#include "nrf52840_bitfields.h"
#include "nrf52840.h"
#include "Radio.h"

#define LAST_RF_CHANNEL 125   // 125 correspond a 2525 MHz.

static unsigned char ad[5] = {0x00, 0x00, 0x41, 0x52, 0x63};

// Function that swaps the bits within each byte in a uint32. Used to convert from nRF24L type addressing to nRF51 type addressing
static unsigned int bytewise_bit_swap(unsigned int inp)
{
    inp = (inp & 0xF0F0F0F0) >> 4U | (inp & 0x0F0F0F0F) << 4U;
    inp = (inp & 0xCCCCCCCC) >> 2U | (inp & 0x33333333) << 2U;
    return (inp & 0xAAAAAAAA) >> 1U | (inp & 0x55555555) << 1U;
}

static void (*Frame_Received_Callback)(void);

void Radio_Init(Radio_Config Config, void* p_data_byte_aligned, void (*p_Frame_Received_Function)(void))
{
	Frame_Received_Callback = p_Frame_Received_Function;
	
	// Disable Radio
	NRF_RADIO->TASKS_DISABLE = 1U;
	while(!NRF_RADIO->EVENTS_DISABLED);
	NRF_RADIO->EVENTS_DISABLED = 0U;

	// Radio config
	NRF_RADIO->TXPOWER   = RADIO_TXPOWER_TXPOWER_Pos4dBm;
	NRF_RADIO->FREQUENCY = LAST_RF_CHANNEL - Config.Channel;
	NRF_RADIO->MODE      = (Config.BitRate << RADIO_MODE_MODE_Pos);

	// Radio address config
	NRF_RADIO->PREFIX0 = (unsigned int)bytewise_bit_swap(Config.Adresses[0]); // Prefix byte of address 0 converted to nRF24L series format

	NRF_RADIO->BASE0 = bytewise_bit_swap(Config.Adresses[1] << 24U | ad[2] << 16U | ad[3] << 8U | ad[4]);  // Base address for prefix 0 converted to nRF24L series format

	NRF_RADIO->TXADDRESS   = 0x00UL;  // Set device address 0 to use when transmitting
	NRF_RADIO->RXADDRESSES = 0x01UL;  // Enable device address 0 to use to select which addresses to receive

	// Packet configuration
	NRF_RADIO->PCNF0 = (0U << RADIO_PCNF0_S0LEN_Pos) | (0U << RADIO_PCNF0_LFLEN_Pos) | (0U << RADIO_PCNF0_S1LEN_Pos); //format sb
	NRF_RADIO->PCNF1 = (RADIO_PCNF1_WHITEEN_Disabled        << RADIO_PCNF1_WHITEEN_Pos) |
										 (RADIO_PCNF1_ENDIAN_Big              << RADIO_PCNF1_ENDIAN_Pos)  |
										 ((sizeof(ad) - 1)										<< RADIO_PCNF1_BALEN_Pos)   |
										 (Config.FrameSize										<< RADIO_PCNF1_STATLEN_Pos) |
										 (Config.FrameSize										<< RADIO_PCNF1_MAXLEN_Pos);

	// CRC Config
	NRF_RADIO->CRCCNF = (RADIO_CRCCNF_LEN_Two << RADIO_CRCCNF_LEN_Pos); // Number of checksum bits
	if ((NRF_RADIO->CRCCNF & RADIO_CRCCNF_LEN_Msk) == (RADIO_CRCCNF_LEN_Two << RADIO_CRCCNF_LEN_Pos))
	{
		NRF_RADIO->CRCINIT = 0xFFFFUL;   // Initial value
		NRF_RADIO->CRCPOLY = 0x11021UL;  // CRC poly: x^16 + x^12^x^5 + 1
	}
	else if ((NRF_RADIO->CRCCNF & RADIO_CRCCNF_LEN_Msk) == (RADIO_CRCCNF_LEN_One << RADIO_CRCCNF_LEN_Pos))
	{
		NRF_RADIO->CRCINIT = 0xFFUL;   // Initial value
		NRF_RADIO->CRCPOLY = 0x107UL;  // CRC poly: x^8 + x^2^x^1 + 1
	}
	
	// Pointer Config
	NRF_RADIO->PACKETPTR = (unsigned int)p_data_byte_aligned;
	
	// Shorts Config
	NRF_RADIO->SHORTS = (RADIO_SHORTS_READY_START_Enabled << RADIO_SHORTS_READY_START_Pos) | 
											(RADIO_SHORTS_END_DISABLE_Enabled << RADIO_SHORTS_END_DISABLE_Pos);
	
	// Interrupt Source
	NRF_RADIO->INTENCLR = 0xFFFFFFFF;
	NRF_RADIO->INTENSET = (RADIO_INTENSET_END_Set << RADIO_INTENSET_END_Pos);
}

void Radio_Start(void)
{
	NRF_RADIO->TASKS_RXEN = 1;
}

void Radio_Stop(void)
{
	NRF_RADIO->TASKS_DISABLE = 1;
}

unsigned int Radio_Get_CRC_Status(void)
{
	return NRF_RADIO->CRCSTATUS;
}

void RADIO_IRQHandler(void)
{
	NRF_RADIO->EVENTS_END = 0;	// we enter this isr only when a frame is received and radio is disabled at the same time
	if(Frame_Received_Callback != 0) Frame_Received_Callback();
}	
