#include "Implant.hpp"
#include "RTC.h"
#include "IRQ_Priority.h"
#include "NVIC.h"
#include "SampleExtractor.hpp"
#include "ports.h"
#include <cstring>
// DEBUG
#include "HSIS.h"
#include "GPIO.h"

#define ADDRESS	{0x00,0x00,0x41,0x52,0x63}
#define ADDRESS_SZ	5
#define B_PREAMBLE	1
#define B_S0	0
#define B_LENGTH	0
#define B_S1	0
#define B_CRC	2
#define EXTRA_TO_READY_US	400 //us
#define PULSE_PERIOD_MS		500
#define RAMP_UP_TIME			140 //us
#define RSYNC_PERIOD			1000 //ms
#define QOS_HIGH	100 // unit: %
#define QOS_LOW	0 // unit: %
#define NB_FRM_LOST 5 // max nb of lost frames to pass on scan mode

//#define USE_TIMER_INSTEAD_OF_RTC

ImplantSignalConfig TImplant::Configuration;
Implant_Status_t TImplant::Status = Implant_Standby;
unsigned int TImplant::nb_PAD = 0;
Implant_Config_t TImplant::Implant_Config;
Wait_Time_t TImplant::Wait_Times;
unsigned short TImplant::IndexRead = 0;
unsigned char TImplant::Implant_Data[IMPLANT_BUF_SIZE];
double TImplant::datarate[3] = {1.0, 2.0, 0.25};
bool TImplant::sample_timeout_flag = false;

static unsigned int round(double number) { return (number >= 0) ? (unsigned int)(number + 0.5) : (unsigned int)(number - 0.5); }

static unsigned int conv_ms_register(unsigned short ms) { return (unsigned int)round(32.768*(double)ms); }

static unsigned int conv_us_register(unsigned short us) { return (unsigned int)round(0.032768*(double) us); }

void TImplant::StartConfig(unsigned char* cfg_ptr)
{
	unsigned char temp, read_idx = 0;
	
	// -- Variable calculation
	Status = Implant_Scan;
	Implant_Config.radio_config.FrameSize = cfg_ptr[read_idx++];
	Implant_Config.RF_period = cfg_ptr[read_idx++];
	Implant_Config.radio_config.Channel = cfg_ptr[read_idx++];
	Implant_Config.radio_config.Adresses[0] = cfg_ptr[read_idx++];
	Implant_Config.radio_config.Adresses[1] = cfg_ptr[read_idx++];
	temp = cfg_ptr[read_idx++];
	Implant_Config.ACQ_period = cfg_ptr[read_idx++];
	Implant_Config.QOS_enable = (bool)((temp & 0x08)>>3);
	Implant_Config.radio_config.BitRate = ((temp & 0x10) == 0x10) ? _250KBPS : _1MBPS;
	PopulateSignals(cfg_ptr + read_idx);
	CalculateWaitTimes();

	// -- Init peripherals
	Radio_Init(Implant_Config.radio_config, Implant_Data, Frame_Received);
	IRQ_SET(IRQ_Radio, Radio_PRIO);
	IRQ_CLEAR_PENDING(IRQ_Radio);
	#ifdef USE_TIMER_INSTEAD_OF_RTC
	Timer_Config timer_cfg = {
		.Channel = ImplantTimer,
		.Bitmode = Bits32,
		.Prescaler = 0x0F
	};
	Timer_Setup(timer_cfg);
	IRQ_SET(IRQ_Timer1, Radio_RTC_PRIO);
	#else
	RTC_Config rtc_cfg ={
		.channel = ImplantRTC,
		.prescaler = RTC_32k768Hz,
		.irq = RTC_ISR_Compare0
	};
	RTC_Setup(rtc_cfg, RTC_Tick);
	IRQ_SET(IRQ_RTC, Radio_RTC_PRIO);
	#endif
	// -- Start peripherals
	StartRTC(Wait_Times.SCAN);
	Radio_Start();
	
	io_set(DEBUG1_Port, DEBUG1_Pin); // DEBUG // breakpoint here to fix rtc
}

#define POPULATE_SIGNAL(sig) \
	ts = cfg_ptr[read_idx++]; \
	if(ts > 0) \
	{ \
		Configuration.sig.SamplesCountPerFrame = (unsigned char) (Configuration.Ts_Slow / ts); \
		Configuration.sig.Enable = true; \
	} \
	else Configuration.sig.Enable = false; \
	Configuration.sig.Resolution = cfg_ptr[read_idx++];

void TImplant::PopulateSignals(unsigned char* cfg_ptr)
{
	unsigned char ts, read_idx = 0;
	Configuration.Ts_Slow = cfg_ptr[read_idx++];
	POPULATE_SIGNAL(iFast1);
	POPULATE_SIGNAL(iFast2);
	POPULATE_SIGNAL(iFast3);
	POPULATE_SIGNAL(iFast4);
	POPULATE_SIGNAL(iT);
	POPULATE_SIGNAL(iAcc);
	POPULATE_SIGNAL(iVbat);
	Configuration.QoS.Enable = true;
	Configuration.QoS.SamplesCountPerFrame = 1;
	Configuration.QoS.Resolution = 8;
}

void TImplant::CalculateWaitTimes(void)
{
	unsigned int frame_time = CalculateFrameTime();
	int AcqSeqCount = (Implant_Config.RF_period / Implant_Config.ACQ_period);
	Wait_Times.NL = ((AcqSeqCount * conv_ms_register(Implant_Config.ACQ_period)) - conv_us_register(frame_time)) - 1;	// AcqSeqCount * ACQ_periode is larger than true RF_period
	Wait_Times.NL_PAD = ((AcqSeqCount * conv_ms_register(Implant_Config.ACQ_period)) - conv_us_register(2 * frame_time)) - 1;
	Wait_Times.L = (conv_us_register(2U * frame_time)) - 1;
	Wait_Times.PULSE = conv_ms_register(PULSE_PERIOD_MS) - 1;
	Wait_Times.SCAN = ((AcqSeqCount * conv_ms_register(Implant_Config.ACQ_period))) + conv_us_register(frame_time) - 1;	// JXU : + frame time

#ifdef USE_TIMER_INSTEAD_OF_RTC
	// adapt to 1MHz timer
	const unsigned int Old_freq = 32768;
	const unsigned int New_freq = 1000000;
	Wait_Times.NL = Wait_Times.NL * New_freq / Old_freq;
	Wait_Times.NL_PAD = Wait_Times.NL_PAD * New_freq / Old_freq;
	Wait_Times.L = Wait_Times.L * New_freq / Old_freq;
	Wait_Times.PULSE = Wait_Times.PULSE * New_freq / Old_freq;
	Wait_Times.SCAN = Wait_Times.SCAN * New_freq / Old_freq;
#endif
}

unsigned int TImplant::CalculateFrameTime(void)
{
	unsigned int time;
	unsigned short size =(B_PREAMBLE + ADDRESS_SZ + B_S0 + B_LENGTH + B_S1 + Implant_Config.radio_config.FrameSize + B_CRC);
	time = size * 8 / datarate[Implant_Config.radio_config.BitRate]; //all byte / datarate 
	time = (unsigned int)(time * 1.1); //xrp janv-2017: add margin to the listening time to avoid missing RF frame
	time += EXTRA_TO_READY_US + RAMP_UP_TIME;
	return time;
}

void TImplant::RTC_Tick(void)
{
	io_tog(DEBUG2_Port, DEBUG2_Pin); // DEBUG
	switch(Status)
	{	
		default : break;
		case Implant_Scan :
			Status = Implant_NoScan;	
			Radio_Stop();
			StartRTC(Wait_Times.SCAN);
			break;
		case Implant_NoScan :
			Status = Implant_Scan;	
			Radio_Start();
			StartRTC(Wait_Times.SCAN);
			break;
		case Implant_Listening :
			Radio_Stop();
			sample_timeout_flag = true; // GAT : OOS interference case see Defect#9660
			if(++nb_PAD < NB_FRM_LOST) // Les 4 premières fois
			{
				Status = Implant_NoListening_PAD;	
				StartRTC(Wait_Times.NL_PAD);
			}
			else // à la 5ème on sleep
			{
				Status = Implant_NoScan;
				nb_PAD = 0;
				StartRTC(Wait_Times.PULSE);
			}
			break;
		case Implant_NoListening :	
		case Implant_NoListening_PAD :
			Status = Implant_Listening;
			Radio_Start();
			StartRTC(Wait_Times.L);
			break;
	}
}


static volatile unsigned char DEBUG[50];
static volatile unsigned char DEBUG_index = 0;

void TImplant::Frame_Received(void)
{
	if(DEBUG_index == sizeof(DEBUG))
	{
		DEBUG[0] = 0;
	}
	else
	{
		DEBUG[DEBUG_index++] = Implant_Data[0];
	}
	io_tog(DEBUG1_Port, DEBUG1_Pin); // DEBUG
	nb_PAD = 0; // reset lost frames count
	StartRTC(Wait_Times.NL); // whenever a frame is received, we pass to NL mode
	Status = Implant_NoListening;
	if(Radio_Get_CRC_Status() == 1)
	{
		IndexRead = 1;
		ProcessFrame();
		Add_QoS_Sample(true);
	}
	else
	{
		Add_QoS_Sample(false);
	}
}

void TImplant::Add_QoS_Sample(bool data)
{
   if(Configuration.QoS.SamplesCount < Configuration.QoS.SamplesCountPerFrame)
	{
		Configuration.QoS.Data[Configuration.QoS.DataWriteIndex++] = data ? QOS_HIGH : QOS_LOW;
		if(Configuration.QoS.DataWriteIndex >= Configuration.QoS.SamplesCountPerFrame) Configuration.QoS.DataWriteIndex = 0;
		Configuration.QoS.SamplesCount++;
	}
	else // Ecriture plus rapide que la lecture (quartz Tx moins rapide)
	{
		// On ne fait rien
	}
}

void TImplant::ProcessFrame(void)
{
   SampleExtractor::Reset();
   ProcessSignal(&Configuration.iFast1);
	ProcessSignal(&Configuration.iFast2);
	ProcessSignal(&Configuration.iFast3);
	ProcessSignal(&Configuration.iFast4);
	ProcessSignal(&Configuration.iT);
	ProcessSignal(&Configuration.iAcc);
	ProcessSignal(&Configuration.iVbat);
}

void TImplant::ProcessSignal(ImplantSignal* signal)
{
	if(signal->Enable)
	{
		for(int i = 0; i < signal->SamplesCountPerFrame; i++)
		{
			if(signal->SamplesCount < signal->SamplesCountPerFrame)
			{
				signal->Data[signal->DataWriteIndex++] = SampleExtractor::Extract(Implant_Data, IMPLANT_BUF_SIZE, &IndexRead, signal->Resolution);
				if(signal->DataWriteIndex >= signal->SamplesCountPerFrame) signal->DataWriteIndex = 0;
				signal->SamplesCount++;
			}
			else // Ecriture plus rapide que la lecture (quartz Tx moins rapide)
			{
				// On doit continuer � lire la trame pour ne pas que le signal suivant r�cup�re le derni�r �chantillon de ce signal
				(void)SampleExtractor::Extract(Implant_Data, IMPLANT_BUF_SIZE, &IndexRead, signal->Resolution);
			}
		}
	}
}

void TImplant::GetSample(unsigned char* data_ptr, ImplantSignalName signal_name)
{
	unsigned short sample;
	switch(signal_name)
	{
		case iFast1:
		sample = GetSigSample(&Configuration.iFast1); break;
		case iFast2:
		sample = GetSigSample(&Configuration.iFast2); break;
		case iFast3:
		sample = GetSigSample(&Configuration.iFast3); break;
		case iFast4:
		sample = GetSigSample(&Configuration.iFast4); break;
		case iT:
		sample = GetSigSample(&Configuration.iT); break;
		case iAcc:
		sample = GetSigSample(&Configuration.iAcc); break;
		case iVbat:
		sample = GetSigSample(&Configuration.iVbat); break;
		case QoS:
		sample = GetQoSSample(); break;
	}
	memcpy(data_ptr, &sample, 2U);
}

unsigned short TImplant::GetSigSample(ImplantSignal* signal)
{
	unsigned short sample;
	if(signal->SamplesCount == 0) return signal->LastSample;
	sample = signal->Data[signal->DataReadIndex++];
	if(signal->DataReadIndex >= signal->SamplesCountPerFrame) signal->DataReadIndex = 0;
	signal->SamplesCount--;
	signal->LastSample = sample;
	return sample;
}

unsigned short TImplant::GetQoSSample(void)
{
	static bool Pad_QoS_Loss = true; // ignore QoS loss the first time we run out of samples
	if(Configuration.QoS.SamplesCount == 0)
	{
		if(sample_timeout_flag) // We trully timed out on implant frame read
		{
			sample_timeout_flag = false; // reset flag
			Pad_QoS_Loss = false; // do not ignore QoS loss
			return QOS_LOW;
		}
		else // We lost a frame because of speed difference
		{
			if(Pad_QoS_Loss)
			{
				Pad_QoS_Loss = false; // Pad just for the first loss
				return Configuration.QoS.LastSample;
			}
			else
				return QOS_LOW;
		}
	}
	Pad_QoS_Loss = true;
	Configuration.QoS.SamplesCount--;
	Configuration.QoS.LastSample = Configuration.QoS.Data[0];
	return Configuration.QoS.Data[0];
}

void TImplant::StartRTC(unsigned int time)
{
	IRQ_CLEAR_PENDING(IRQ_Radio);
	#ifdef USE_TIMER_INSTEAD_OF_RTC
	Timer_Compare_Config timer_channel_cfg = {
		.Channel = ImplantTimer,
		.Compare_Channel = ImplantTimer_Compare_Channel,
		.Shortcut = Stop,
		.CompareValue = time,
		.Callback = RTC_Tick
	};
	Timer_Set_Compare(timer_channel_cfg);
	Timer_Start(ImplantTimer);
	#else
	RTC_Start(ImplantRTC, time);
	#endif
}

