// SHT3X, Temp. Rel. Hum. Transducer Middleware | gat 23/06/23
#include "TRH.hpp"
#include "TWIM.h"
#include "FreeRTOS.h"
#include "task.h"
#include "semphr.h"
#include <cstring> // memcpy
#include "ports.h"
#include "HSIS.h"
#include "GPIO.h"

#define TWI_Address 0x44 // 0x44 ou 0x45

#define Read_MSB 0x27
#define Read_LSB_High	0x37
#define Read_LSB_Medium	0x21
#define Read_LSB_Low		0x2A

#define Break_MSB		0x30
#define Break_LSB		0x93

static bool TRH_running = false;

void TRH::Enable(void)
{
	TRH_running = true;
	//io_set(EN_TRH_Port, EN_TRH_Pin);
}

void TRH::Disable(void)
{
	TRH_running = false;
	//io_clr(EN_TRH_Port, EN_TRH_Pin);
}

#define IdleSleepTime 100U
#define MutexWaitingTime 10U // Time to wait before giving up on waiting for the port mutex to be released
#define AfterRequestWaitingTime 20U // Time to wait after tx before rx
#define CriticalPortUnlockTimeout 20U // This timeout makes sure we dont retain the port mutex forever
#define ReadbackLength 6U // byte size we read from transducer

//static unsigned char IsAcquisitionEnabled = 0;
static unsigned char data_buffer[ReadbackLength];
extern SemaphoreHandle_t TWI0_Mutex;

void TRH::StartBlocking(void)
{
	unsigned char request[] = {Read_MSB, Read_LSB_High};
	if(xSemaphoreTake(TWI0_Mutex, MutexWaitingTime) == pdTRUE) // Reserve port
	{
		TWIM_Tx(TRH_TWIM_Channel, TWI_Address, sizeof(request), request, xTaskGetCurrentTaskHandle()); // Send a conversion request
		(void)ulTaskNotifyTake(pdTRUE, CriticalPortUnlockTimeout); // Sleep until end of Tx
	} 
}

void TRH::Start(void)
{
	unsigned char request[] = {Read_MSB, Read_LSB_High};
	if(xSemaphoreTake(TWI0_Mutex, MutexWaitingTime) == pdTRUE) // Reserve port
	{
		TWIM_Tx(TRH_TWIM_Channel, TWI_Address, sizeof(request), request, NULL); // Send a conversion request
	} 
}

void TRH::ReadData(void)
{
	TWIM_Status port_status;
	unsigned char local_data_buffer[ReadbackLength];
	if(xSemaphoreTake(TWI0_Mutex, MutexWaitingTime) == pdTRUE) // Reserve port
	{
		TWIM_Rx(TRH_TWIM_Channel, TWI_Address, ReadbackLength, local_data_buffer, xTaskGetCurrentTaskHandle()); // Read data
		(void)ulTaskNotifyTake(pdTRUE, CriticalPortUnlockTimeout); // Sleep until end of Rx
		port_status = TWIM_ReadStatus(TRH_TWIM_Channel); // Get port status
		if(!port_status.AddressNACK) 
			memcpy(data_buffer, local_data_buffer, ReadbackLength); // Validate local data if an address NACK wasnt issued
	}
}

unsigned char* TRH::GetDataPtr()
{
	return data_buffer;
}

typedef union
{
	unsigned int _int16;
	unsigned char _byte[2];
}sample_type;

#define CONVERT_TEMP

//#define TEST1

void TRH::GetTemperatureSample(unsigned char* data_ptr)
{
	#ifdef TEST1
	unsigned char dummyTEST[] = {0x78,0x9A};
	memcpy(data_ptr, dummyTEST, sizeof(dummyTEST));
	#else
	#ifdef CONVERT_TEMP
	sample_type sample;
	double tmp;
	sample._byte[0] = data_buffer[1];
	sample._byte[1] = data_buffer[0];
	tmp = sample._int16;
	tmp *= 17500;
	tmp /= 65535;
	tmp -= 4500;
	sample._int16 = tmp;
	data_ptr[0] = sample._byte[1];
	data_ptr[1] = sample._byte[0];
	#else
	memcpy(data_ptr, data_buffer, 2U); // RAW SAMPLE
	#endif // CONVERT_TEMP
	#endif // TEST1
}

#define CONVERT_HUMIDITY
//#define TEST2

void TRH::GetHumiditySample(unsigned char* data_ptr)
{
	#ifdef TEST2
	unsigned char dummyTEST[] = {0xBC,0xDE};
	memcpy(data_ptr, dummyTEST, sizeof(dummyTEST));
	#else
	#ifdef CONVERT_HUMIDITY
	sample_type sample;
	double tmp;
	sample._byte[0] = data_buffer[4];
	sample._byte[1] = data_buffer[3];
	tmp = sample._int16;
	tmp *= 10000;
	tmp /= 65535;
	sample._int16 = tmp;
	data_ptr[0] = sample._byte[1];
	data_ptr[1] = sample._byte[0];
	#else
	memcpy(data_ptr, data_buffer + 3, 2U); // RAW SAMPLE
	#endif
	#endif // TEST
}

typedef enum
{
	No_Detection,
	Detected,
	Undetected
} Detection_Status_t;

static Detection_Status_t detection_status = No_Detection;

bool TRH::Detect(void)
{
	//bool TemporaryEnable = !TRH_running;
	//if(TemporaryEnable) io_set(EN_TRH_Port, EN_TRH_Pin);
	//vTaskDelay(1);
	Break();
	//if(TemporaryEnable) io_clr(EN_TRH_Port, EN_TRH_Pin);
	TWIM_Status status = TWIM_ReadStatus(TRH_TWIM_Channel);
	bool ret = status.AddressNACK == 0;
	detection_status = ret ? Detected : Undetected;
	return ret;
}

void TRH::Break(void)
{
	unsigned char request[] = {Break_MSB, Break_LSB};
	if(xSemaphoreTake(TWI0_Mutex, MutexWaitingTime) == pdTRUE) // Reserve port
	{
		TWIM_Tx(TRH_TWIM_Channel, TWI_Address, sizeof(request), request, xTaskGetCurrentTaskHandle()); // Send a conversion request
		(void)ulTaskNotifyTake(pdTRUE, CriticalPortUnlockTimeout); // Sleep until end of Tx
	} 
}

bool TRH::IsDetected(void)
{
	if(detection_status == No_Detection) Detect();
	return detection_status == Detected;
}
